#define F_CPU 8000000UL

#include <stdlib.h>
#include <stdio.h>
#include <avr/io.h>
#include <util/delay.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/sleep.h>
#include <avr/sfr_defs.h>
#include <avr/pgmspace.h>
#include <avr/eeprom.h>
#include <util/atomic.h>
#include "dogm128.h"


// vstupy
#define BUTT1 PINB, 3
#define BUTT2 PINB, 4

// vystupy
#define A PORTA, 7
#define B PORTA, 6
#define C PORTA, 5

uint8_t amux[8]={0x0,0x80,0x40,0xC0,0x20,0xA0,0x60,0xE0};

//#define CS PORTB, 2
//#define A0 PORTB, 1
//#define RST PORTB, 0

enum {X1C,X1B,X1A,X2C,X2B,X2A,X3C};

// napeti baterii * 10
//uint8_t batt[7] = {177,177,177,123,123,177,177}; // testovací hodnoty
uint8_t batt[7] = {0,0,0,0,0,0,0};
uint16_t batt_p[7] = {0,0,0,0,0,0,0};
uint16_t batt_o[7] = {0,0,0,0,0,0,0};
uint16_t batt_oo[7] = {0,0,0,0,0,0,0};

// celkove napeti
uint16_t sbatt=0;

// korekce ofsetu
//uint8_t batt_ofe[7]= {1,1,1,1,1,1,1};

// gain error
uint8_t batt_gae[7]={96, 96, 96, 96, 96, 96, 96};
uint8_t EEMEM e_batt_gae[7]={96, 96, 96, 96, 96, 96, 96};

// minimalni napeti baterie (male - small) - nastavitelne
uint8_t EEMEM e_min_vs = 123;
uint8_t min_vs=123;

// max. napeti baterie (male) - nastavitelne
uint8_t EEMEM e_max_vs = 177;
uint8_t max_vs=177;

// minimalni napeti baterie (velke - big) - nastavitelne
uint8_t EEMEM e_min_vb = 123;
uint8_t min_vb=123;

// max. napeti baterie (velke) - nastavitelne
uint8_t EEMEM e_max_vb = 177;
uint8_t max_vb=177;

// prumerne napeti baterie
uint8_t av_batt=0;

// hodnota ve V*10 pro varovani o prilisnem vybiti jedne baterky oproti prumeru
uint8_t EEMEM e_av_war=25;
uint8_t av_war=25;

// minimalni celkova hodnota napeti (varovani) - nastavitelne
uint16_t EEMEM e_ov_war=861;
uint16_t ov_war=861;

// koeficienty pro velkou "tuzkovku" - pocitaji se
float kt=0.0, qt=0.0;

// koef. pro malou tuzkovku
float ks=0.0, qs=0.0;


volatile int16_t adc_val;

char str[30]={};

// priznak aktivovane editace nejakeho param.
volatile uint8_t sc_edit=0;

volatile uint8_t b2count=0;

//volatile uint8_t state=0;


/*ISR(SPI_STC_vect) {

	SETBIT(SPCR,MSTR); // bug - pin SS

}*/

ISR(ADC_vect)
{
	adc_val = ADCW;
}

ISR(TIMER1_COMPA_vect) {



}

// vypocet koef pro ukazatele miry nabiti
void calc_kq() {

	// koeficienty pro velkou tuzkovku
	kt = 7*max_vb-7*min_vb;
	kt = -60.0/kt;
	//kt = -0.15873;

	qt=(60*min_vb) / (max_vb - min_vb);
	qt += 60;
	//qt = 196.66;

	// koef. pro male tuzkovky
	ks = max_vs-min_vs;
	ks = -18.0/ks;

	qs=(18*min_vs) / (max_vs - min_vs);
	qs += 18;

}


void amux_sel(uint8_t chan) {

	PORTA &= 0x1F; // vynulovani pinu A B C

	PORTA |= amux[chan]; // nastaveni spravne kombinace

}

// spusteni AD prevodu - noise reduction mode
void run_adc() {

	set_sleep_mode(SLEEP_MODE_ADC);

	sleep_mode();

}

// zobrazeni "tuzkovek"
void disp_total() {

	uint8_t i=0;

	int16_t wb = kt*sbatt+qt; // vypocet bileho mista u velke tuzkovky
	if (wb>59) wb = 59;
	if (wb<0) wb=0;

	int16_t y = 63-wb-9; // omezeni na vhodne hodnoty
	if (y<0) y=0;
	if (y>(63-10)) y=63-10;

	dog_StartPage();

	do {

		dog_ClrBox(1,0,127,63);

		// velka baterie vlevo
		dog_SetBox(1,0,20,61); // obrys velke baterie
		dog_SetBox(9,62,13,63); // cudlik nahore

		if (wb>0) dog_ClrBox(2,60-wb,19,60); // bila cast, y: 1-60

		sprintf(str,"%i.%iV",sbatt/10,sbatt%10);
		dog_DrawStr(24, y, font_9x18, str); // napis se posouva podle nabiti baterky

		//mala baterie vpravo
		for(i=0;i<=6;i++) {

			int16_t wbs = ks*batt[i]+qs; // bila cast malych baterek
			if (wbs>18) wbs = 18;
			if (wbs<0) wbs=0;

			dog_SetBox(80,1+9*i,100,8+9*i); // obrys male baterie 1
			dog_SetBox(101,4+9*i,102,5+9*i); // cudlik
			if (wbs>0) dog_ClrBox(99-wbs,2+9*i,99,7+9*i); // bila cast

			if (batt[i]!=255) sprintf(str,"%i.%iV",batt[i]/10,(batt[i]%10));
			else sprintf(str,"?.??V");
			//sprintf(str,"%i",batt_gae[i]);

			dog_DrawStr(104, 1+9*i, font_5x7, str);

		} // for

		if (sbatt<ov_war && ov_war != 7*min_vb) { // varovani: celkove napeti < ov_war
		// varovani se nezobrazi pokud ov_war==min_vb

			//dog_ClrBox(5,22,5+dog_GetStrWidth(font_9x18,"batteries"),35+14);
			dog_ClrBox(5,22,5+80,35+14);
			dog_DrawStr(5, 37, font_9x18, "Low");
			dog_DrawStr(5, 22, font_9x18, "batteries");
			//dog_XorBox(5,22,5+dog_GetStrWidth(font_9x18,"batteries"),35+14);
			dog_XorBox(5,22,5+80,35+14);

				}


	} while( dog_NextPage() );


}


void start_timer1() {

	TCNT1=0;

	TCCR1B = (0<<CS12)|(1<<CS11)|(1<<CS10)|(1<<WGM12);

}

void stop_timer1() {

	TCCR1B = (0<<CS12)|(0<<CS11)|(0<<CS10)|(1<<WGM12);

}

uint8_t check_butt1() {

	uint8_t i=0;

	// s - indikace, ze tlacitko je stisknuto dele, v co se pocita jak dlouho
	volatile static uint8_t s=0,co=0;

	if (C_CHECKBIT(BUTT1)) s=0; // tl. neni stisknute

	if (s==1) {

		// kdyz je tl. stisk. dyl nez dve sec, pocitame, ze je to schvalne
		if (co<40) co++;
		else return 1;

	} else co = 0;

	if (s==0) {

		while (!C_CHECKBIT(BUTT1) && i<10) {

			i++;
			_delay_us(50);

		};

		if (i==10) {

			s = 1;
			return 1;

		}
		else return 0;
	}

	return 0;

}

uint8_t check_butt2() {

	uint8_t i=0;

	// osklivy "hack" kvuli tomu, ze prave tlacitko je na pinu SS,
	// ktery kdyz je nast jako vstup
	// prepina pri stazeni do nuly SPI do rezimu slave

	CLEARBIT(SPCR,SPE); // vypnuti SPI
	CLEARBIT(DDRB,DDB4);

	// s - indikace, ze tlacitko je stisknuto dele, v co se pocita jak dlouho
	static uint8_t s=0,co=0;

	if (C_CHECKBIT(BUTT2)) s=0; // tl. neni stisknute

	if (s==1) {

		// kdyz je tl. stisk. dyl nez dve sec, pocitame ze je to schvalne
		if (co<40) co++;
		else {

			SETBIT(DDRB,DDB4);
			SETBIT(SPCR,SPE); // zapnuti SPI (bug - pin SS)
			return 1;

		}

	} else co = 0;


	if (s==0) {
		while (!C_CHECKBIT(BUTT2) && i<10) { // osetreni zakmitu

			i++;
			_delay_us(50);

		};

		SETBIT(DDRB,DDB4);
		SETBIT(SPCR,SPE); // zapnuti SPI (bug - pin SS)

		if (i==10) {

			s = 1;
			return 1;

		}
		else return 0;

	}

	SETBIT(DDRB,DDB4);
	SETBIT(SPCR,SPE); // zapnuti SPI (bug - pin SS)

	return 0;

}

volatile uint8_t mux=0;

// obrazovka D_MAIN
void d_main() {

	//static uint8_t mux=0;
	static uint16_t war_int=0;
	uint32_t sum=0;

	static uint8_t wr=0;

	// zmereni napeti vsech baterii
	for (mux=0; mux<=6; mux++) {

		amux_sel(mux);

		_delay_ms(10);

		run_adc();

		if (adc_val>511) {

			if (adc_val>900) {
				batt_p[mux]=0;
			} else batt_p[mux]=2550;

		} else {

			uint32_t batt_a = (((uint32_t)adc_val*7120)/2048);

			//batt_a = (double)((double)(((batt_a * batt_gae[mux])/100) - (double)batt_ofe[mux]) + 0.5);
			batt_a = (double)((double)((batt_a * batt_gae[mux])/100) - 1.0 + 0.5);

			if (batt_a<50) batt_a = 0; /* mene nez 0.5 V ignorujeme */

			batt_p[mux] = (batt_a + batt_o[mux] + batt_oo[mux])/3;

			batt_oo[mux] = batt_o[mux];
			batt_o[mux] = batt_a;

		}

		if (batt_p[mux]!=2550) sum += batt_p[mux]; // 255 je spec. hodnota - nezapocitat do prumeru

		batt[mux] = (batt_p[mux]+5)/10;

	} // for

	// prumerne napeti baterie
	av_batt = (uint8_t)((sum +35) / 70);

	// ulozeni vysledku do glob prom
	sbatt = (sum) / 10;

	for (mux=X1C; mux<=X3C; mux++) {

		if ((av_batt>av_war) && batt[mux] < av_batt-av_war) wr++; // pocet bat ktere jsou pod prumerem

	} // for


	war_int++;
	if (war_int==(60)) war_int=0;

	if (wr==0 || war_int<(30)) {

		disp_total(); // obsluha displeje

	} else {

		wr = 0;

		dog_StartPage();

		do {

			dog_SetBox(1,0,127,63);
			dog_ClrBox(2,1,126,62);

			dog_DrawStr(5,52 , font_5x7, "Unbalanced discharge!");

			snprintf(str,30,"Uav: %d.%dV",av_batt/10,av_batt%10);
			dog_DrawStr(10,40, font_5x7, str);

			uint8_t i=0;

			for (mux=X1C; mux<=X3C; mux++) {

				if (batt[mux] < av_batt-av_war) {

					snprintf(str,30,"Ub%d is %d.%dV bellow.",mux+1,(av_batt-batt[mux])/10,(av_batt-batt[mux])%10);
					dog_DrawStr(10,30-i*8 , font_5x7, str);

					i++;

				} // if


			} // for

		} while( dog_NextPage());

	} // if wr

}

// obrazovka D_MINMAX
void d_minmax() {

	//char str[30];

	enum {NO_EDIT, MAXS, MINS, MAXB, MINB}; // s - small, b - big

	volatile static uint8_t state = NO_EDIT;

	if (check_butt2()) {

		//bcount++;

		//if (bcount>=4) {
			state++;

		//	bcount = 0;
			b2count = 0;

			if (state>MINB) {

				state = NO_EDIT;

				// nove nastaveni zapiseme do EEPROM
				eeprom_write_byte(&e_min_vs,min_vs);
				eeprom_write_byte(&e_max_vs,max_vs);
				eeprom_write_byte(&e_min_vb,min_vb);
				eeprom_write_byte(&e_max_vb,max_vb);

				// prepocitame koeficienty pro tuzkovky
				calc_kq();

			}

		//} // if bcount

	} // if checkbutt

	if (state!=NO_EDIT) sc_edit=1;
	else sc_edit=0;

	if (state==MAXS) { // max. male baterky

		if (check_butt1()) {

			if (b2count<8*5*5) b2count++;

			if (b2count==8*5*5) max_vs += 10;
			else max_vs++;


		} else b2count = 0;

		if (max_vs>177) max_vs=min_vs;

	}

	if (state==MINS) { // min. male baterky

		if (check_butt1()) {

			if (b2count<8*5*5) b2count++;

			if (b2count==8*5*5) min_vs += 10;
			else min_vs++;

		}
		if (min_vs>max_vs) min_vs=0;

	}

	if (state==MAXB) { // max. velke baterky

		if (check_butt1()) {

			if (b2count<8*5*5) b2count++;

			if (b2count==8*5*5) max_vb += 10;
			else max_vb++;

			} else b2count = 0;

		if (max_vb>177) max_vb=min_vb;

	}

	if (state==MINB) { // min. velke baterky

			if (check_butt1()) {

				if (b2count<8*5*5) b2count++;

				if (b2count==8*5*5) min_vb += 10;
				else min_vb++;

			}
		if (min_vb>max_vb) min_vb=0;

		}

	dog_StartPage();

	do {

		dog_ClrBox(0,0,128,64);

		dog_DrawStr(5,52 , font_5x7, "Level settings");

		uint8_t w = 110;

		// mal baterie
		snprintf(str,30,"Max s. voltage: %i.%dV",max_vs/10,max_vs%10);
		dog_DrawStr(10,40 , font_5x7, str);
		if (state==MAXS) {

			//uint8_t w = dog_GetStrWidth(font_5x7,str);

			dog_XorBox(9,40,9+w+2,42+5); // zvyrazneni editovaneho policka

		}

		snprintf(str,30,"Min s. voltage: %i.%dV",min_vs/10,min_vs%10);
		dog_DrawStr(10,30 , font_5x7, str);
		if (state==MINS) {

			//uint8_t w = dog_GetStrWidth(font_5x7,str);

			dog_XorBox(9,30,9+w+2,32+5); // zvyrazneni

		}

		// velk baterie
		snprintf(str,30,"Max b. voltage: %i.%dV",(uint16_t)(7*max_vb)/10,(uint16_t)(7*max_vb)%10);
		dog_DrawStr(10,20 , font_5x7, str);
		if (state==MAXB) {

			//uint8_t w = dog_GetStrWidth(font_5x7,str);

			dog_XorBox(9,20,9+w+2,22+5); // zvyrazneni editovaneho policka

		}

		snprintf(str,30,"Min b. voltage: %i.%dV",(uint16_t)(7*min_vb)/10,(uint16_t)(7*min_vb)%10);
		dog_DrawStr(10,10 , font_5x7, str);
		if (state==MINB) {

			//uint8_t w = dog_GetStrWidth(font_5x7,str);

			dog_XorBox(9,10,9+w+2,12+5); // zvyrazneni

		}


	} while( dog_NextPage());
}

// obrazovka D_WAR
void d_war() {

	//char str[30];

	enum {NO_EDIT, AV_WAR, OV_WAR};

		volatile static uint8_t state=NO_EDIT;
		//state = NO_EDIT;

		//static uint8_t b2count=0;

		if (check_butt2()) {

			//bcount++;

			//if (bcount>=4) {
				state++;

			//	bcount = 0;
				b2count = 0;

				if (state>OV_WAR) {

					state = NO_EDIT;

					// nove nastaveni zapsat do EEPROM
					eeprom_write_byte(&e_av_war,av_war);
					eeprom_write_word(&e_ov_war,ov_war);

				}

			//} // if bcount

		} // if checkbutt

		if (state!=NO_EDIT) sc_edit=1;
		else sc_edit=0;

		if (state==AV_WAR) {

			if (check_butt1()) {

				if (b2count<8*5*5) b2count++;

				if (b2count==8*5*5) av_war += 10;
				else av_war++;


			} else b2count = 0;

			if (av_war>60) av_war=1;

		}

		if (state==OV_WAR) {

				if (check_butt1()) {

					if (b2count<8*5*5) b2count++;

					if (b2count==8*5*5) ov_war += 10;
					else ov_war++;

				}

				if (ov_war>(7*max_vb)) ov_war=7*min_vb;

			}

	dog_StartPage();

	do {

		uint8_t w = 110;

		dog_ClrBox(0,0,128,64);

		dog_DrawStr(5,52 , font_5x7, "Warning settings");

		snprintf(str,30,"Inequal dischar.: %i.%iV",av_war/10,av_war%10);
		dog_DrawStr(10,40 , font_5x7, str);
		if (state==AV_WAR) {

			//uint8_t w = dog_GetStrWidth(font_5x7,str);

			dog_XorBox(9,40,9+w+2,42+5); // zvyrazneni

		}

		snprintf(str,30,"Minimal overall");
		dog_DrawStr(10,30 , font_5x7, str);

		if (ov_war!=7*min_vb) {

			snprintf(str,30,"bat. volt.   : %i.%iV",ov_war/10,ov_war%10);
			dog_DrawStr(10,20 , font_5x7, str);

		} else { // vypnuti varovani LOW BATT -> napis DIS

			snprintf(str,30,"baterries volt. : DIS");
			dog_DrawStr(10,20 , font_5x7, str);

		}

		if (state==OV_WAR) {

			//uint8_t w = dog_GetStrWidth(font_5x7,str);

			dog_XorBox(9,20,9+w+2,32+5); // zvyrazneni

				}




	} while( dog_NextPage());
}

// obrazovka D_GAIN
void d_gain() {

	enum {NO_EDIT, B1, B2, B3, B4, B5, B6, B7};

	volatile static uint8_t state=NO_EDIT;

	volatile uint8_t mux = 1;

	if (check_butt2()) {

			state++;

			if (state>B7) {

				state = NO_EDIT;

				// nove nastaveni zapiseme do EEPROM
				eeprom_write_block(&batt_gae,&e_batt_gae,7);

			}



	} // if checkbutt

	if (state!=NO_EDIT) sc_edit=1;
	else sc_edit=0;

	if (state!=NO_EDIT) {

		mux = state-1;

		if (check_butt1()) {

			batt_gae[mux] += 1;

		}

		if (batt_gae[mux]>120) batt_gae[mux] = 80;

	}


	dog_StartPage();

	do {

		dog_ClrBox(0,0,128,64);

		dog_DrawStr(5,52 , font_5x7, "Gain error compensation");


		for (int8_t i=6;i>=0;i--) {

			snprintf(str,30,"B%i: %i%d%%",i+1,batt_gae[i]/10,batt_gae[i]%10);
			dog_DrawStr((i>2)?10:60,40+(10*((i-6)%4)) , font_5x7, str);
			if (state==i+1) {
				dog_XorBox((i>2)?10:60,40+(10*((i-6)%4)),((i>2)?10:60)+40,9+40+(10*((i-6)%4))); // zvyrazneni editovaneho policka
			}

		} // for


	} while( dog_NextPage());
}

int main (void) {

	// nastaveni A,B,C jako vystup
	DDRA = (1<<DDA5)|(1<<DDA6)|(1<<DDA7);

	// nast. I/O k disp. jako vystup
	DDRB = (1<<DDB0)|(1<<DDB1)|(1<<DDB2)|(0<<DDB3)|(1<<DDB4)|(1<<DDB5)|(1<<DDB7);
	// pull-upy u tlacitek
	PORTB = (1<<PORTB3)|(1<<PORTB4);

	// ADC externi ref., dif ADC1-0 10x
	ADMUX = (0<<REFS1)|(0<<REFS0)|(0<<MUX4)|(1<<MUX3)|(0<<MUX2)|(0<<MUX1)|(1<<MUX0);

	// povoleni ADC, 8 MHz/64
	ADCSRA = (1<<ADEN)|(1<<ADPS2)|(1<<ADPS1)|(0<<ADPS0)|(1<<ADIE);

	// nastaveni SPI, master, fclk/4
	SPCR = 0;
	SPCR =  (1<<SPE) | (1<<MSTR)|(0<<SPR1)|(0<<SPR0)|(1<<CPOL)|(1<<CPHA)|(0<<SPIE);

	// nastaveni ct1: CTC, 64x, 100 Hz
	OCR1A = 624;
	//TCCR1B = (1<<CS12)|(0<<CS11)|(1<<CS10)|(1<<WGM12);
	TIMSK = (1<<OCIE1A);

	// nacteni hodnot z EEPROM
	//if (eeprom_read_byte(&e_min_v)!=0xff) {

	min_vs = eeprom_read_byte(&e_min_vs);
	max_vs = eeprom_read_byte(&e_max_vs);
	min_vb = eeprom_read_byte(&e_min_vb);
	max_vb = eeprom_read_byte(&e_max_vb);
	av_war = eeprom_read_byte(&e_av_war);

	eeprom_read_block(&batt_gae,&e_batt_gae,7);

	//}

	// spocitat koef pro tuzkovky
	calc_kq();

	// inicializace displeje
	dog_Init();

	dog_ClrBox(0,0,128,64);

	sei();

	// MAIN - jedna velka a sedm malych tuzkovek
	// MINMAX - nastaveni minima a maxima baterii
	// WAR - nastaveni urovni pro varovani
	enum {D_MAIN,D_MINMAX,D_WAR,D_GAIN};

	uint8_t disp=D_MAIN;

	while(1) { // 100 Hz

		// stavovy automat - jednotlive obrazovky
		switch (disp) {

			case D_MAIN: d_main(); break;

			case D_MINMAX: d_minmax(); break;

			case D_WAR: d_war(); break;

			case D_GAIN: d_gain(); break;


		}; // switch

		// prepnuti obrazovky
		if (!sc_edit && check_butt1()) {

			//while (check_butt1());

			if (disp<D_GAIN)	disp++;
			else disp=D_MAIN;

		} // if check butt1


		// uspani na pul sekundy
		start_timer1();
		CLEARBIT(ADCSRA, ADIE); // aby ADC neprobouzel procesor ze spanku
		set_sleep_mode(SLEEP_MODE_IDLE);
		sleep_mode();
		SETBIT(ADCSRA, ADIE);
		stop_timer1();


	} // while


} // main
